local updateFrame = CreateFrame("Frame")
local minFallHeight, maxFallHeight = 13, 70
local fallDistance = maxFallHeight - minFallHeight
local ACCEL = 19 / 2
local CATFORM = 3

local _, playerClass = UnitClass("player")

local _G = getfenv(0)
local GetInventoryItemLink = _G.GetInventoryItemLink
local GetItemInfo = _G.GetItemInfo
local GetContainerNumSlots = _G.GetContainerNumSlots
local GetContainerItemLink = _G.GetContainerItemLink
local GetContainerItemInfo = _G.GetContainerItemInfo
local PickupContainerItem = _G.PickupContainerItem
local ContainerIDToInventoryID = _G.ContainerIDToInventoryID
local AutoEquipCursorItem = _G.AutoEquipCursorItem
local GetSpellInfo = _G.GetSpellInfo
local IsFalling = _G.IsFalling
local UnitHealth, UnitHealthMax = _G.UnitHealth, _G.UnitHealthMax
local GetTime = _G.GetTime
local wipe, select, pairs, ipairs = _G.wipe, _G.select, _G.pairs, _G.ipairs

local isFloating = false
local hasFelineGrace = false

local function isNormalBag(bagId)
	if bagId == 0 or bagId == -1 then return true end
	local link = GetInventoryItemLink("player", ContainerIDToInventoryID(bagId))
	if not link then return false end
	local linkId = select(3, link:find("item:(%d+)"))
	if not linkId then return false end
	local bagType = select(7, GetItemInfo(linkId))
	if bagType and bagType == INVTYPE_BAG then return true end
	return false
end

-- Adapted from ClosetGnome
local slotLocks = {}
local function locateFreeSlot()
	for theBag = NUM_BAG_FRAMES, 0, -1 do
		if isNormalBag(theBag) then
			local numSlot = GetContainerNumSlots(theBag)
			for theSlot = 1, numSlot, 1 do
				if not slotLocks[theBag .. "-" .. theSlot] then
					local texture = GetContainerItemInfo(theBag, theSlot)
					if not texture then
						return theBag, theSlot
					end
				end
			end
		end
	end
	return nil
end

local gearPriority = {16, 17, 18, 1, 5, 7, 3, 10, 8, 6, 9}
AieeeeeeItems = {}

local function unequipItems()
	for i, slot in ipairs(gearPriority) do
		local rangedType = nil
		if slot == 18 then
			local link = GetInventoryItemLink("player", slot)
			if link then
				local id = link:find("item:(%d+)")
				if id then
					rangedType = select(7, GetItemInfo(id))
				end
			end
		end
		if rangedType ~= "INVTYPE_THROWN" and rangedType ~= "INVTYPE_RELIC" then
			PickupInventoryItem(slot)
			local toBag, toSlot = locateFreeSlot()
			if toBag ~= nil then
				slotLocks[toBag .. "-" .. toSlot] = true
				AieeeeeeItems[slot] = GetInventoryItemLink("player", slot)
				PickupContainerItem(toBag, toSlot)
			else
				AutoEquipCursorItem()
				break
			end
		end
	end
	slotLocks = wipe(slotLocks)
end

local function unequipGearsToSavePreciousGolds()
	ChatFrame1:AddMessage("|cffff0000Aieeeeee!|r It looks like you are going to fall to your death! At least you'll be naked.")
	unequipItems()
end

local function restoreGearAfterHorribleMessySplattyDeath()
	if UnitIsDead("player") or UnitIsGhost("player") then return end
	for slot, link in pairs(AieeeeeeItems) do
		for theBag = NUM_BAG_FRAMES, 0, -1 do
			local numSlot = GetContainerNumSlots(theBag)
			local equipped = false
			for theSlot = 1, numSlot, 1 do
				if GetContainerItemLink(theBag, theSlot) == link then
					PickupContainerItem(theBag, theSlot)
					AutoEquipCursorItem()
					equipped = true
					break
				end
			end
			if equipped then break end
		end
		if GetInventoryItemLink("player", slot) == link then
			AieeeeeeItems[slot] = nil
		end
	end
end

local falling = 0
local startedUnequip = false
local fallEndTime = 0
local function checkFalling(self, t)
	if not isFloating and IsFalling() and not startedUnequip then		
		falling = falling + t
		if playerClass == "DRUID" and hasFelineGrace then
			if GetShapeshiftForm() == CATFORM then
				minFallHeight, maxFallHeight = 29, 87
			else
				minFallHeight, maxFallHeight = 13, 70
			end
		end
		
		local distance = (ACCEL * falling * falling) - minFallHeight
		if (distance / fallDistance) >= (UnitHealth("player") / UnitHealthMax("player")) then
			startedUnequip = true
			unequipGearsToSavePreciousGolds()
		end
	elseif not IsFalling() or isFloating then
		startedUnequip = false
		if falling > 0 then
			fallEndTime = GetTime()
		end
		falling = 0
		if GetTime() - fallEndTime > 1 then
			for k, v in pairs(AieeeeeeItems) do
				restoreGearAfterHorribleMessySplattyDeath()
				break
			end
		end
	end
end

-- Safe fall adds 17 yards
-- Glyph of Safe Fall adds 10 yards
-- Non-rogues can fall 13 yards before damage
-- Rogues can fall 39 yards with Safe Fall + glyph before damage

local mod = {}

function mod:PLAYER_ENTERING_WORLD()
	local safeFall = GetSpellInfo(1860)
	local felineGrace = GetSpellInfo(20719)
	hasFelineGrace = GetSpellInfo(felineGrace)
	if GetSpellInfo(safeFall) then
		minFallHeight, maxFallHeight = 29, 87
	end
	
	for i = 1, GetNumGlyphSockets() do
		local _, _, spellID = GetGlyphSocketInfo(i)
		if spellID == 58033 then	-- Glyph of safe fall	
			minFallHeight, maxFallHeight = 39, 97
		end
	end
	
	fallDistance = maxFallHeight - minFallHeight
end

function mod:PLAYER_ENTERING_WORLD()
	self:PLAYER_ALIVE()
end

function mod:PLAYER_UNGHOST()
	self:PLAYER_ALIVE()
end

function mod:PLAYER_ALIVE()
	falling = 0
	restoreGearAfterHorribleMessySplattyDeath()
	updateFrame:SetScript("OnUpdate", checkFalling)
end

function mod:PLAYER_REGEN_ENABLED()
	self:PLAYER_ALIVE()
end

local slowfallBuffs = {
	1706,		-- levitate
	32028,		-- tears of the goddess
	30424,		-- ghostwalk
	24927,		-- halloween candy
	33382,		-- jump-o-tron
	16717,		-- launch
	55416,		-- slowfall
	24742,		-- magic wings
	42867,		-- magic wings
	16593,		-- noggenfogger
	37897,		-- parachute
	44795,		-- parachute,
	45472,		-- parachute
	50550,		-- parachute
	50554,		-- parachute
	53208,		-- parachute
	54168,		-- parachute
	54649,		-- parachute
	61242,		-- parachute
	61243,		-- parachute
	61302,		-- parachute
	61360,		-- parachute
	130,		-- slow fall
	12438,		-- slow fall, parachute cloak
	50085,		-- slow fall
	50237,		-- slow fall
	58441,		-- snowfall lager	
}

function mod:UNIT_AURA(event, unit)
	if unit ~= "player" then return end
	isFloating = false
	if IsFalling() then
		for _, id in ipairs(slowfallBuffs) do
			local name = GetSpellInfo(id)
			if UnitAura("player", name) then
				isFloating = true
			end
		end
	end
end

local function onEvent(self, event, ...)
	if mod[event] then
		mod[event](mod, event, ...)
	end
end

updateFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
updateFrame:RegisterEvent("PLAYER_UNGHOST")
updateFrame:RegisterEvent("PLAYER_ALIVE")
updateFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
updateFrame:RegisterEvent("UNIT_AURA")
updateFrame:SetScript("OnEvent", onEvent)
